@@ -1,7 +1,7 @@
 /*
  * GRSPW2 SpaceWire Kernel Library Interface
  *
- * Copyright (c) 2010 Aeroflex Gaisler AB
+ * Copyright (c) 2016 Cobham Gaisler AB
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -76,16 +76,16 @@ struct grspw_pkt;
 #define TXPKT_FLAG_NOCRC_LENe 0x0000000e
 #define TXPKT_FLAG_NOCRC_LENf 0x0000000f
 
-/* Marks if packet was transmitted or not */
-#define TXPKT_FLAG_TX 0x8000
-
 #define TXPKT_FLAG_INPUT_MASK (TXPKT_FLAG_NOCRC_MASK | TXPKT_FLAG_IE | \
 				TXPKT_FLAG_HCRC | TXPKT_FLAG_DCRC)
 
-/* Link Error */
-#define TXPKT_FLAG_LINKERR 0x4000
+/* Marks if packet was transmitted or not */
+#define TXPKT_FLAG_TX 0x4000
 
-#define TXPKT_FLAG_OUTPUT_MASK (TXPKT_FLAG_LINKERR)
+/* Link Error */
+#define TXPKT_FLAG_LINKERR 0x8000
+
+#define TXPKT_FLAG_OUTPUT_MASK (TXPKT_FLAG_TX | TXPKT_FLAG_LINKERR)
 
 /*** RX Packet Flags ***/
 
@@ -107,6 +107,15 @@ struct grspw_pkt;
 
 #define RXPKT_FLAG_OUTPUT_MASK (RXPKT_FLAG_TRUNK | RXPKT_FLAG_DCRC | \
 				RXPKT_FLAG_HCRC | RXPKT_FLAG_EEOP)
+
+/*** General packet flag options ***/
+
+/* Translate Hdr and/or Payload address */
+#define PKT_FLAG_TR_DATA_RESERVED 0x1000
+#define PKT_FLAG_TR_HDR_RESERVED 0x2000
+/* All General options */
+#define PKT_FLAG_MASK_RESERVED 0x3000
+
 #endif
 /* GRSPW RX/TX Packet structure.
  *
@@ -126,14 +135,14 @@ struct grspw_pkt;
  *   type (RX/TX). See XXPKT_FLAG_* options above.
  */
 struct grspw_pkt {
-	struct grspw_pkt *next;
-	unsigned int pkt_id;	/* User assigned ID */
-	unsigned short flags;	/* RX/TX Options */
+	struct grspw_pkt *next;	/* Next packet in list. NULL if last packet */
+	unsigned int pkt_id;	/* User assigned ID (not touched by driver) */
+	unsigned short flags;	/* RX/TX Options and status */
 	unsigned char reserved;	/* Reserved, must be zero */
-	unsigned char hlen;	/* Length of Header Buffer */
+	unsigned char hlen;	/* Length of Header Buffer (only TX) */
 	unsigned int dlen;	/* Length of Data Buffer */
 	void *data;	/* 4-byte or byte aligned depends on HW */
-	void *hdr;	/* 4-byte or byte aligned depends on HW */
+	void *hdr;	/* 4-byte or byte aligned depends on HW (only TX) */
 };
 
 /* GRSPW SpaceWire Packet List */
@@ -147,8 +156,9 @@ typedef enum {
 	SPW_LS_ERRRST = 0,
 	SPW_LS_ERRWAIT = 1,
 	SPW_LS_READY = 2,
-	SPW_LS_STARTED = 3,
-	SPW_LS_RUN = 4
+	SPW_LS_CONNECTING = 3,
+	SPW_LS_STARTED = 4,
+	SPW_LS_RUN = 5
 } spw_link_state_t;
 
 /* Address Configuration */
@@ -188,7 +198,9 @@ struct grspw_core_stats {
 	int err_eeop;
 	int err_addr;
 	int err_parity;
+	int err_disconnect;
 	int err_escape;
+	int err_wsync; /* only in GRSPW1 */
 };
 
 /* grspw_link_ctrl() options */
@@ -196,20 +208,47 @@ struct grspw_core_stats {
 #define LINKOPTS_DISABLE	0x0001
 #define LINKOPTS_START		0x0002
 #define LINKOPTS_AUTOSTART	0x0004
-#define LINKOPTS_DIS_ONERR	0x0008
-#define LINKOPTS_IRQ		0x0200	/* Enable Error Link IRQ */
-#define LINKOPTS_MASK		0x020f /* All above options */
+#define LINKOPTS_DIS_ONERR	0x0008	/* Disable DMA transmitter on link error
+					 * Controls LE bit in DMACTRL register.
+					 */
+#define LINKOPTS_DIS_ON_CE	0x0020000/* Disable Link on Credit error */
+#define LINKOPTS_DIS_ON_ER	0x0040000/* Disable Link on Escape error */
+#define LINKOPTS_DIS_ON_DE	0x0080000/* Disable Link on Disconnect error */
+#define LINKOPTS_DIS_ON_PE	0x0100000/* Disable Link on Parity error */
+#define LINKOPTS_DIS_ON_WE	0x0400000/* Disable Link on write synchonization
+					  * error (GRSPW1 only)
+					  */
+#define LINKOPTS_DIS_ON_EE	0x1000000/* Disable Link on Early EOP/EEP error*/
+
+/*#define LINKOPTS_TICK_OUT_IRQ	0x0100*//* Enable Tick-out IRQ */
+#define LINKOPTS_EIRQ		0x0200	/* Enable Error Link IRQ */
+
+#define LINKOPTS_MASK		0x15e020f/* All above options */
+#define LINKOPTS_MASK_DIS_ON	0x15e0000/* All disable link on error options
+					  * On a certain error the link disable
+					  * bit will be written and the work
+					  * task will call dma_stop() for all
+					  * channels.
+					  */
+
+#define LINKSTS_CE		0x002	/* Credit error */
+#define LINKSTS_ER		0x004	/* Escape error */
+#define LINKSTS_DE		0x008	/* Disconnect error */
+#define LINKSTS_PE		0x010	/* Parity error */
+#define LINKSTS_WE		0x040	/* Write synchonization error (GRSPW1 only) */
+#define LINKSTS_IA		0x080	/* Invalid address */
+#define LINKSTS_EE		0x100	/* Early EOP/EEP */
+#define LINKSTS_MASK		0x1de
 
 /* grspw_tc_ctrl() options */
-#define TCOPTS_EN_RX	0x0001
-#define TCOPTS_EN_RXIRQ	0x0002	/* Tick-Out IRQ */
+#define TCOPTS_EN_RXIRQ	0x0001	/* Tick-Out IRQ */
 #define TCOPTS_EN_TX	0x0004
+#define TCOPTS_EN_RX	0x0008
 
 /* grspw_rmap_ctrl() options */
 #define RMAPOPTS_EN_RMAP	0x0001
 #define RMAPOPTS_EN_BUF		0x0002
 
-
 /* grspw_dma_config.flags options */
 #define DMAFLAG_NO_SPILL	0x0001	/* See HW doc DMA-CTRL NS bit */
 #define DMAFLAG_RESV1		0x0002	/* HAS NO EFFECT */
@@ -217,18 +256,22 @@ struct grspw_core_stats {
 #define DMAFLAG_STRIP_PID	0x0008	/* See HW doc DMA-CTRL SP bit */
 #define DMAFLAG_RESV2		0x0010	/* HAS NO EFFECT */
 #define DMAFLAG_MASK	(DMAFLAG_NO_SPILL|DMAFLAG_STRIP_ADR|DMAFLAG_STRIP_PID)
+/* grspw_dma_config.flags misc options (not shifted internally) */
+#define DMAFLAG2_TXIE	0x00100000	/* See HW doc DMA-CTRL TI bit. 
+					 * Used to enable TX DMA interrupt
+					 * when tx_irq_en_cnt=0.
+					 */
+#define DMAFLAG2_RXIE	0x00200000	/* See HW doc DMA-CTRL RI bit.
+					 * Used to enable RX DMA interrupt
+					 * when rx_irq_en_cnt=0.
+					 */
+#define DMAFLAG2_MASK	(DMAFLAG2_TXIE | DMAFLAG2_RXIE)
 
 struct grspw_dma_config {
-	int flags;
-
+	int flags;		/* DMA config flags, see DMAFLAG1&2_* options */
 	int rxmaxlen;		/* RX Max Packet Length */
 	int rx_irq_en_cnt;	/* Enable RX IRQ every cnt descriptors */
 	int tx_irq_en_cnt;	/* Enable TX IRQ every cnt descriptors */
-
-#if 0
-	int rx_desc_cnt;	/* Number of RX Descriptor */
-	int tx_desc_cnt;	/* Number of TX Descriptor */
-#endif
 };
 
 /* Statistics per DMA channel */
@@ -251,12 +294,16 @@ struct grspw_dma_stats {
 	int tx_sched_cnt_min;	/* Minimum number of packets in TX SCHED Q */
 	int tx_sched_cnt_max;	/* Maximum number of packets in TX SCHED Q */
 	int sent_cnt_max;	/* Maximum number of packets in TX SENT Q */
+	int tx_work_cnt;	/* Times the work thread processed TX BDs */
+	int tx_work_enabled;	/* No. RX BDs enabled by work thread */
 
 	int ready_cnt_min;	/* Minimum number of packets in RX READY Q */
 	int ready_cnt_max;	/* Maximum number of packets in RX READY Q */
 	int rx_sched_cnt_min;	/* Minimum number of packets in RX SCHED Q */
 	int rx_sched_cnt_max;	/* Maximum number of packets in RX SCHED Q */
 	int recv_cnt_max;	/* Maximum number of packets in RX RECV Q */
+	int rx_work_cnt;	/* Times the work thread processed RX BDs */
+	int rx_work_enabled;	/* No. RX BDs enabled by work thread */
 };
 
 #ifdef __KERNEL__
@@ -271,12 +318,11 @@ extern void grspw_initialize_user(
 	);
 extern int grspw_dev_count(void);
 extern void *grspw_open(int dev_no);
-extern void grspw_close(void *d);
+extern int grspw_close(void *d);
 extern void grspw_hw_support(void *d, struct grspw_hw_sup *hw);
 extern void grspw_stats_read(void *d, struct grspw_core_stats *sts);
 extern void grspw_stats_clr(void *d);
 
-
 /* Set and Read current node address configuration. The dma_nacfg[N] field
  * represents the configuration for DMA Channel N.
  *
@@ -293,7 +339,11 @@ extern spw_link_state_t grspw_link_state(void *d);
  *  bits 7..0  : Clock Div RUN (only run-state)
  *  bits 15..8 : Clock Div During Startup (all link states except run-state)
  */
-extern void grspw_link_ctrl(void *d, int *options, int *clkdiv);
+extern void grspw_link_ctrl(void *d, int *options, int *stscfg, int *clkdiv);
+/* Read the current value of the status register */
+extern unsigned int grspw_link_status(void *d);
+/* Clear bits in the status register */
+extern void grspw_link_status_clr(void *d, unsigned int clearmask);
 
 /*** Time Code Interface ***/
 /* Generate Tick-In (increment Time Counter, Send Time Code) */
@@ -329,7 +379,7 @@ extern int grspw_port_active(void *d);
 
 /*** DMA Interface ***/
 extern void *grspw_dma_open(void *d, int chan_no);
-extern void grspw_dma_close(void *c);
+extern int grspw_dma_close(void *c);
 
 extern int grspw_dma_start(void *c);
 extern void grspw_dma_stop(void *c);
@@ -356,6 +406,11 @@ extern void grspw_dma_stop(void *c);
  *  B. grspw_dma_tx_send(opts=1)
  *
  * NOTE: the TXPKT_FLAG_TX flag must not be set.
+ *
+ * Return Code
+ *  -1   Error
+ *  0    Successfully added pkts to send/sched list
+ *  1    DMA stopped. No operation.
  */
 extern int grspw_dma_tx_send(void *c, int opts, struct grspw_list *pkts, int count);
 
@@ -374,21 +429,39 @@ extern int grspw_dma_tx_send(void *c, int opts, struct grspw_list *pkts, int cou
  * is to call:
  *  A. grspw_dma_tx_reclaim(opts=2) (Skip step 3)
  *  B. grspw_dma_tx_send(opts=1) (Skip step 1)
+ *
+ * Return Code
+ *  -1   Error
+ *  0    Successful. pkts list filled with all packets from sent list
+ *  1    Same as 0, but indicates that DMA stopped
  */
 extern int grspw_dma_tx_reclaim(void *c, int opts, struct grspw_list *pkts, int *count);
 
 /* Get current number of Packets in respective TX Queue. */
-extern void grspw_dma_tx_count(void *c, int *send, int *sched, int *sent);
+extern void grspw_dma_tx_count(void *c, int *send, int *sched, int *sent, int *hw);
 
-/* Block until ready_cnt or fewer packets are Queued in "Send and Scheduled" Q,
+#define GRSPW_OP_AND 0
+#define GRSPW_OP_OR 1
+/* Block until send_cnt or fewer packets are Queued in "Send and Scheduled" Q,
  * op (AND or OR), sent_cnt or more packet "have been sent" (Sent Q) condition
  * is met.
  * If a link error occurs and the Stop on Link error is defined, this function
  * will also return to caller.
+ * The timeout argument is used to return after timeout ticks, regardless of
+ * the other conditions. If timeout is zero, the function will wait forever
+ * until the condition is satisfied.
  *
- * NOTE: if IRQ of TX descriptors are not enabled, this will hang infinitely
+ * NOTE: if IRQ of TX descriptors are not enabled conditions are never
+ *       checked, this may hang infinitely unless a timeout has been specified
+ *
+ * Return Code
+ *  -1   Error
+ *  0    Returing to caller because specified conditions are now fullfilled
+ *  1    DMA stopped
+ *  2    Timeout, conditions are not met
+ *  3    Another task is already waiting. Service is Busy.
  */
-extern int grspw_dma_tx_wait(void *c, int send_cnt, int op, int sent_cnt);
+extern int grspw_dma_tx_wait(void *c, int send_cnt, int op, int sent_cnt, long timeout);
 
 /* Get received RX packet buffers that has previously been scheduled for 
  * reception with grspw_dma_rx_prepare().
@@ -405,6 +478,11 @@ extern int grspw_dma_tx_wait(void *c, int send_cnt, int op, int sent_cnt);
  * packet buffers for future receive, is to call:
  *  A. grspw_dma_rx_recv(opts=2, &recvlist) (Skip step 3)
  *  B. grspw_dma_rx_prepare(opts=1, &freelist) (Skip step 1)
+ *
+ * Return Code
+ *  -1   Error
+ *  0    Successfully filled pkts list with packets from recv list.
+ *  1    DMA stopped
  */
 extern int grspw_dma_rx_recv(void *c, int opts, struct grspw_list *pkts, int *count);
 
@@ -423,21 +501,37 @@ extern int grspw_dma_rx_recv(void *c, int opts, struct grspw_list *pkts, int *co
  * packet buffers for future receive, is to call:
  *  A. grspw_dma_rx_recv(opts=2, &recvlist) (Skip step 3)
  *  B. grspw_dma_rx_prepare(opts=1, &freelist) (Skip step 1)
+ *
+ * Return Code
+ *  -1   Error
+ *  0    Successfully added packet buffers from pkt list into the ready queue
+ *  1    DMA stopped
  */
 extern int grspw_dma_rx_prepare(void *c, int opts, struct grspw_list *pkts, int count);
 
 /* Get current number of Packets in respective RX Queue. */
-extern void grspw_dma_rx_count(void *c, int *ready, int *sched, int *recv);
+extern void grspw_dma_rx_count(void *c, int *ready, int *sched, int *recv, int *hw);
 
 /* Block until recv_cnt or more packets are Queued in RECV Q, op (AND or OR), 
  * ready_cnt or fewer packet buffers are available in the "READY and Scheduled" Q,
  * condition is met.
  * If a link error occurs and the Stop on Link error is defined, this function
  * will also return to caller, however with an error.
+ * The timeout argument is used to return after timeout ticks, regardless of
+ * the other conditions. If timeout is zero, the function will wait forever
+ * until the condition is satisfied.
  *
- * NOTE: if IRQ of RX descriptors are not enabled, this will hang infinitely
+ * NOTE: if IRQ of RX descriptors are not enabled conditions are never
+ *       checked, this may hang infinitely unless a timeout has been specified
+ *
+ * Return Code
+ *  -1   Error
+ *  0    Returing to caller because specified conditions are now fullfilled
+ *  1    DMA stopped
+ *  2    Timeout, conditions are not met
+ *  3    Another task is already waiting. Service is Busy.
  */
-extern int grspw_dma_rx_wait(void *c, int recv_cnt, int op, int ready_cnt);
+extern int grspw_dma_rx_wait(void *c, int recv_cnt, int op, int ready_cnt, long timeout);
 
 
 extern int grspw_dma_config(void *c, struct grspw_dma_config *cfg);
