@@ -1,5 +1,7 @@
 /*
- * Aeroflex Gaisler GRSPW2 user space access Driver.
+ * Cobham Gaisler GRSPW user space access Driver.
+ *
+ * Copyright (c) 2016 Cobham Gaisler AB
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -43,6 +45,13 @@
 #error GRSPW: not enough CHAR device nodes
 #endif
 
+#ifndef init_MUTEX
+#define init_MUTEX(a) sema_init(a, 1);
+#endif
+#ifndef init_MUTEX_LOCKED
+#define init_MUTEX_LOCKED(a) sema_init(a, 0);
+#endif
+
 /* Which Memory MAP device node that should be used. Currently hardcoded.
  * should be configurable per GRSPW Device during runtime.
  */
@@ -73,9 +82,10 @@ struct grspwu_priv {
 	void *_pkt_tx_structs;	/* Base address of packet structures */
 	void *_pkt_rx_structs;	/* Base address of packet structures */
 
-	struct semaphore sem_pktstrs;	/* Free Packet structures Semaphore */
-	struct grspw_list tx_pktstrs;	/* Free TX Packet structures */
-	struct grspw_list rx_pktstrs;	/* Free RX Packet structures */
+	struct semaphore sem_rx_pktstrs;/* Free RX Pkt structures Semaphore */
+	struct semaphore sem_tx_pktstrs;/* Free TX Pkt structure Semaphore */
+	struct grspw_list tx_pktstrs;	/* Free TX Pkt structures list */
+	struct grspw_list rx_pktstrs;	/* Free RX Pkt structures list */
 
 	/* Memory MAP Library instance */
 	struct maplib_drv mmapdrv;
@@ -164,7 +174,8 @@ static int grspwu_open(struct inode *inode, struct file *file)
 	grspw_list_clr(&priv->tx_pktstrs);
 	grspw_list_clr(&priv->rx_pktstrs);
 
-	init_MUTEX(&priv->sem_pktstrs);
+	init_MUTEX(&priv->sem_rx_pktstrs);
+	init_MUTEX(&priv->sem_tx_pktstrs);
 
 	return 0;
 }
@@ -172,6 +183,7 @@ static int grspwu_open(struct inode *inode, struct file *file)
 static int grspwu_release(struct inode *inode, struct file *file)
 {
 	struct grspwu_priv *priv = container_of(inode->i_cdev, struct grspwu_priv, cdev);
+	int i;
 
 	if ( priv->state == STATE_STARTED )
 		grspwu_stop(priv);
@@ -184,6 +196,12 @@ static int grspwu_release(struct inode *inode, struct file *file)
 		priv->_pkt_rx_structs = NULL;
 	}
 	priv->state = STATE_STOPPED;
+	for (i=0; i<priv->chan_cnt; i++) {
+		if (priv->dma[i]) {
+			grspw_dma_close(priv->dma[i]);
+			priv->dma[i] = NULL;
+		}
+	}
 	grspw_close(priv->dh);
 	priv->dh = NULL;
 	file->private_data = NULL;
@@ -284,6 +302,7 @@ int grspwu_config_set(struct grspwu_priv *priv, struct grspw_config *cfg)
 		} else {
 			/* Should be closed */
 			if ( priv->dma[i] ) {
+				grspw_dma_stop(priv->dma[i]);
 				grspw_dma_close(priv->dma[i]);
 				priv->dma[i] = NULL;
 			}
@@ -346,7 +365,7 @@ void grspwu_stats_clr(struct grspwu_priv *priv)
 
 int grspwu_link_ctrl(struct grspwu_priv *priv, struct grspw_link_ctrl *ctrl)
 {
-	int clkdiv;
+	int clkdiv, stscfg;
 
 	/* Check Configuration */
 	if ( ctrl->ctrl & ~LINKOPTS_MASK )
@@ -354,7 +373,7 @@ int grspwu_link_ctrl(struct grspwu_priv *priv, struct grspw_link_ctrl *ctrl)
 
 	/* Configure */
 	clkdiv = ctrl->clkdiv_start<<8 | ctrl->clkdiv_run;
-	grspw_link_ctrl(priv->dh, &ctrl->ctrl, &clkdiv);
+	grspw_link_ctrl(priv->dh, &ctrl->ctrl, &stscfg, &clkdiv);
 
 	return 0;
 }
@@ -385,7 +404,7 @@ void grspwu_link_state_get(struct grspwu_priv *priv, struct grspw_link_state *st
 	/* Get Current Link Configuration */
 	state->link_ctrl = -1;
 	clkdiv = -1;
-	grspw_link_ctrl(priv->dh, &state->link_ctrl, &clkdiv);
+	grspw_link_ctrl(priv->dh, &state->link_ctrl, NULL, &clkdiv);
 	state->clkdiv_start = (clkdiv >> 8) & 0xff;
 	state->clkdiv_run = clkdiv & 0xff;
 
@@ -401,18 +420,18 @@ void grspwu_link_state_get(struct grspwu_priv *priv, struct grspw_link_state *st
 }
 
 /* Set TCTRL and TIMECNT, and Send Time-code */
-void grspwu_tc_send(struct grspwu_priv *priv, int options)
+void grspwu_tc_send(struct grspwu_priv *priv, unsigned int options)
 {
 	int time;
 
 	/* Set TCTRL and TIMECNT? */
-	if ( options & 0x100 ) {
-		time = options & 0xff;
+	if ( options & GRSPW_TC_SEND_OPTION_SET_TCTRL ) {
+		time = options & GRSPW_TC_SEND_OPTION_TCTRL;
 		grspw_tc_time(priv->dh, &time);
 	}
 
 	/* Send Time Code (Generate Tick-In) */
-	if ( options & 0x200 )
+	if ( options & GRSPW_TC_SEND_OPTION_TCTX )
 		grspw_tc_tx(priv->dh);
 }
 
@@ -431,12 +450,14 @@ void grspwu_qpktcnt_read(struct grspwu_priv *priv, struct grspw_qpktcnt *qpktcnt
 				priv->dma[i],
 				&qpktcnt->chan[i].rx_ready,
 				&qpktcnt->chan[i].rx_sched,
-				&qpktcnt->chan[i].rx_recv);
+				&qpktcnt->chan[i].rx_recv,
+				&qpktcnt->chan[i].rx_hw);
 			grspw_dma_tx_count(
 				priv->dma[i],
 				&qpktcnt->chan[i].tx_send,
 				&qpktcnt->chan[i].tx_sched,
-				&qpktcnt->chan[i].tx_sent);
+				&qpktcnt->chan[i].tx_sent,
+				&qpktcnt->chan[i].tx_hw);
 		}
 	}
 }
@@ -589,6 +610,54 @@ void grspwu_unmap(int idx, void *priv)
 	grspwu_stop(priv);
 }
 
+int grspwu_rx_wait(struct grspwu_priv *priv, struct grspw_rx_wait_chan *wait)
+{
+	int rc;
+
+	if (wait->chan >= priv->chan_cnt)
+		return -EINVAL;
+	if (priv->dma[wait->chan] == NULL)
+		return -EBUSY;
+
+	rc = grspw_dma_rx_wait(priv->dma[wait->chan],
+			       wait->recv_cnt, wait->op, wait->ready_cnt,
+			       wait->timeout_ms * HZ/1000);
+	if (rc == -1)
+		return -EAGAIN; /* unknown error */
+	if (rc == 1)
+		return -EIO;
+	else if (rc == 2)
+		return -ETIME;
+	else if (rc == 3)
+		return -EBUSY;
+	else
+		return 0;
+}
+
+int grspwu_tx_wait(struct grspwu_priv *priv, struct grspw_tx_wait_chan *wait)
+{
+	int rc;
+
+	if (wait->chan >= priv->chan_cnt)
+		return -EINVAL;
+	if (priv->dma[wait->chan] == NULL)
+		return -EBUSY;
+
+	rc = grspw_dma_tx_wait(priv->dma[wait->chan],
+			       wait->send_cnt, wait->op, wait->sent_cnt,
+			       wait->timeout_ms * HZ/1000);
+	if (rc == -1)
+		return -EAGAIN; /* unknown error */
+	if (rc == 1)
+		return -EIO;
+	else if (rc == 2)
+		return -ETIME;
+	else if (rc == 3)
+		return -EBUSY;
+	else
+		return 0;
+}
+
 static long grspwu_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
 {
 	struct grspwu_priv *priv;
@@ -698,7 +767,7 @@ static long grspwu_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
 
 	case GRSPW_IOCTL_TC_SEND:
 	{
-		grspwu_tc_send(priv, (int)argp);
+		grspwu_tc_send(priv, (unsigned int)argp);
 		break;
 	}
 
@@ -720,6 +789,26 @@ static long grspwu_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
 		break;
 	}
 
+	case GRSPW_IOCTL_STATUS_READ:
+	{
+		unsigned int status;
+		status = grspw_link_status(priv->dh);
+		if(copy_to_user(argp, &status, sizeof(unsigned int)))
+			return -EFAULT;
+		break;
+	}
+
+	case GRSPW_IOCTL_STATUS_CLR:
+	{
+		unsigned int clr_mask;
+		if ( priv->state == STATE_STARTED)
+			return -EBUSY;
+		if(copy_from_user(&clr_mask, argp, sizeof(unsigned int)))
+			return -EFAULT;
+		grspw_link_status_clr(priv->dh, clr_mask);
+		break;
+	}
+
 	case GRSPW_IOCTL_START:
 	{
 		return grspwu_start(priv, (int)argp);
@@ -731,6 +820,22 @@ static long grspwu_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
 		break;
 	}
 
+	case GRSPW_IOCTL_RX_WAIT:
+	{
+		struct grspw_rx_wait_chan wait;
+		if(copy_from_user(&wait, argp, sizeof(struct grspw_rx_wait_chan)))
+			return -EFAULT;
+		return grspwu_rx_wait(priv, &wait);
+	}
+
+	case GRSPW_IOCTL_TX_WAIT:
+	{
+		struct grspw_tx_wait_chan wait;
+		if(copy_from_user(&wait, argp, sizeof(struct grspw_tx_wait_chan)))
+			return -EFAULT;
+		return grspwu_tx_wait(priv, &wait);
+	}
+	
 	default: return -ENOSYS;
 	}
 
@@ -770,8 +875,12 @@ int grspwu_reclaim(
 	}
 #endif
 	if ( count < 1 ) {
+		/* IMPLEMENT BLOCKING RECLAIM PER DMA CHANNEL HERE.
+		 * The ioctl() implements the TX_WAIT currenlty.
+		 */
+
 		/* If blocking wait until a number of Packets are available */
-#warning IMPLEMENT BLOCKING RECLAIM PER DMA CHANNEL.
+		
 		return 0;
 	}
 
@@ -817,7 +926,7 @@ int grspwu_reclaim(
 
 out:
 	/* Take List lock */
-	while ( down_interruptible(&priv->sem_pktstrs) )
+	while ( down_interruptible(&priv->sem_tx_pktstrs) )
 		;
 
 	/* Queue up packet structs for reuse */
@@ -825,7 +934,7 @@ out:
 		grspw_list_prepend_list(&priv->tx_pktstrs, &sentpkts);
 
 	/* Release Lock */
-	up(&priv->sem_pktstrs);
+	up(&priv->sem_tx_pktstrs);
 
 	return retval;
 }
@@ -863,8 +972,11 @@ int grspwu_recv(
 	}
 #endif
 	if ( count < 1 ) {
+		/* IMPLEMENT BLOCKING RECV PER DMA CHANNEL HERE.
+		 * Currently ioctl() implements the RX_WAIT instead.
+		 */
+
 		/* If blocking wait until a number of Packets are available */
-#warning IMPLEMENT BLOCKING RECV PER DMA CHANNEL.
 		return 0;
 	}
 
@@ -911,14 +1023,14 @@ int grspwu_recv(
 	retval = count;
 out:
 	/* Take List lock */
-	while ( down_interruptible(&priv->sem_pktstrs) )
+	while ( down_interruptible(&priv->sem_rx_pktstrs) )
 		;
 
 	/* Queue up packet structs for reuse */
 	grspw_list_prepend_list(&priv->rx_pktstrs, &recvpkts);
 
 	/* Release Lock */
-	up(&priv->sem_pktstrs);
+	up(&priv->sem_rx_pktstrs);
 
 	return retval;
 }
@@ -947,14 +1059,14 @@ int grspwu_send(struct grspwu_priv *priv, int dmachan,
 	tmp2 = cnt;
 #endif
 
-	while ( down_interruptible(&priv->sem_pktstrs) )
+	while ( down_interruptible(&priv->sem_tx_pktstrs) )
 		;
 
 	/* Take out as many Packet Structs from FREE Packet List as needed */
 	cnt = grspw_list_take_head_list(&priv->tx_pktstrs, &txpkts, cnt);
 
 	/* Release Lock */
-	up(&priv->sem_pktstrs);
+	up(&priv->sem_tx_pktstrs);
 #ifdef GRSPWU_DEBUG
 	tmp = grspw_list_cnt(&txpkts);
 	printk(KERN_DEBUG "grspwu_send(): req %d got %d packets, chain: %d\n", tmp2, count, tmp);
@@ -1053,12 +1165,12 @@ out:
 	/* Return all Packet structures, send none of the read packets due
 	 * to an error. " Rewind all work".
 	 */
-	while ( down_interruptible(&priv->sem_pktstrs) )
+	while ( down_interruptible(&priv->sem_tx_pktstrs) )
 		;
 
 	grspw_list_prepend_list(&priv->tx_pktstrs, &txpkts);
 
-	up(&priv->sem_pktstrs);
+	up(&priv->sem_tx_pktstrs);
 
 	return retval;
 }
@@ -1084,13 +1196,13 @@ int grspwu_prepare(struct grspwu_priv *priv, int dmachan,
 	grspw_list_clr(&readypkts);
 	count = 0;
 
-	while ( down_interruptible(&priv->sem_pktstrs) )
+	while ( down_interruptible(&priv->sem_rx_pktstrs) )
 		;
 
 	/* Take out as many Packet Structs from FREE Packet List as needed */
 	cnt = grspw_list_take_head_list(&priv->rx_pktstrs, &readypkts, cnt);
 
-	up(&priv->sem_pktstrs);
+	up(&priv->sem_rx_pktstrs);
 #ifdef GRSPWU_DEBUG
 	tmp = grspw_list_cnt(&readypkts);
 	printk(KERN_DEBUG "grspwu_prepare(): got %d packets, chain: %d\n", cnt, tmp);
@@ -1178,12 +1290,12 @@ out:
 	/* Return all Packet structures, pepare none of the read packets due
 	 * to an error. " Rewind all work".
 	 */
-	while ( down_interruptible(&priv->sem_pktstrs) )
+	while ( down_interruptible(&priv->sem_rx_pktstrs) )
 		;
 
 	grspw_list_prepend_list(&priv->rx_pktstrs, &readypkts);
 
-	up(&priv->sem_pktstrs);
+	up(&priv->sem_rx_pktstrs);
 
 	return retval;
 }
@@ -1402,6 +1514,7 @@ static void __exit grspwu_exit (void)
 		if ( priv )
 			grspwu_dev_del(i, priv);
 	}
+	grspw_initialize_user(NULL, NULL);
 
 	unregister_chrdev_region(devno, GRSPWU_DEVCNT);
 }
@@ -1409,6 +1522,6 @@ static void __exit grspwu_exit (void)
 module_init(grspwu_init);
 module_exit(grspwu_exit);
 
-MODULE_AUTHOR("Daniel Hellstrom, Aeroflex Gaisler AB.");
-MODULE_DESCRIPTION("Aeroflex Gaisler GRSPW SpaceWire Driver");
+MODULE_AUTHOR("Cobham Gaisler AB.");
+MODULE_DESCRIPTION("Cobham Gaisler GRSPW SpaceWire Driver");
 MODULE_LICENSE("GPL");
