@@ -1,7 +1,7 @@
 /*
  * GRSPW Driver Access Help Library
  *
- * Copyright (c) 2010 Aeroflex Gaisler AB
+ * Copyright (c) 2016 Cobham Gaisler AB
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -36,6 +36,17 @@
 #undef RX_RECV_ADD_END
 #undef TX_RECLAIM_ADD_END
 
+/* In newer versions of GLIBC read() and write() are not accessed directly,
+ * instead some buffer-overlow checking are performed first. The checks does
+ * not work with the GRSPW driver since the driver uses some bits in the
+ * count argument to select which DMA channel is accessed, and which operation
+ * (read incomming packet, read used transmitted packet buffer).
+ */
+ssize_t __read(int fd, void *buf, size_t count);
+ssize_t __write(int fd, const void *buf, size_t count);
+#define READ(fd, buf, count) __read(fd, buf, count)
+#define WRITE(fd, buf, count) __write(fd, buf, count)
+
 /* Used For debug */
 #define PRINT(args...)
 /*#define PRINT(args...) printf(args)*/
@@ -325,6 +336,68 @@ int grspw_stop(struct grspw_dev *dev)
 	return 0;
 }
 
+/* 0xff=TCTRL|TIMECNT, 0x100=Write-Enable, 0x200=Generate-Tick-In */
+int grspw_tc_set(struct grspw_dev *dev, unsigned int options)
+{
+	int result;
+
+	result = ioctl(dev->fd, GRSPW_IOCTL_TC_SEND, options);
+	if ( result < 0 ) {
+		PRINT_ERR("Failed to call TC_SEND for dev %d, errno: %d\n", dev->index, errno);
+		return -1;
+	}
+
+	return 0;
+}
+
+int grspw_tc_read(struct grspw_dev *dev, int *time)
+{
+	int result;
+
+	if (!time)
+		return -1;
+	*time = 0;
+
+	result = ioctl(dev->fd, GRSPW_IOCTL_TC_READ, time);
+	if ( result < 0 ) {
+		PRINT_ERR("Failed to call TC_READ for dev %d, errno: %d\n", dev->index, errno);
+		return -1;
+	}
+
+	return 0;
+}
+
+int grspw_status_read(struct grspw_dev *dev, unsigned int *status)
+{
+	int result;
+
+	if (!status)
+		return -1;
+	*status = 0;
+
+	result = ioctl(dev->fd, GRSPW_IOCTL_STATUS_READ, status);
+	if ( result < 0 ) {
+		PRINT_ERR("Failed to call STATUS_READ for dev %d, errno: %d\n", dev->index, errno);
+		return -1;
+	}
+
+	return 0;
+}
+
+int grspw_link_status_clr(struct grspw_dev *dev, unsigned int clearmask)
+{
+	int result;
+	unsigned int mask = clearmask;
+
+	result = ioctl(dev->fd, GRSPW_IOCTL_STATUS_CLR, &mask);
+	if ( result < 0 ) {
+		PRINT_ERR("Failed to call STATUS_CLR for dev %d, errno: %d\n", dev->index, errno);
+		return -1;
+	}
+
+	return 0;
+}
+
 /* Put Packet Chain 'lst' into the GRSPW driver's SEND List. The packets on
  * the SEND list will at some point make it into a descriptor. What triggers
  * packets to be scheduled are:
@@ -386,7 +459,7 @@ int grspw_tx_send(struct grspw_dev *dev, int dma_chan, struct spwlib_list *lst)
 		 */
 		count = size | GRSPW_WRITE_SEND | 
 			((dma_chan << GRSPW_WRITE_CHAN_BIT) & GRSPW_WRITE_CHAN);
-		result = write(dev->fd, buf, count);
+		result = WRITE(dev->fd, buf, count);
 		if ( result < 0 ) {
 			PRINT_ERR("SpW%d: SEND %d failed: %d\n", dev->index, spwlib_list_cnt(lst), errno);
 			return -1;
@@ -460,7 +533,7 @@ int grspw_tx_reclaim(struct grspw_dev *dev, int dma_chan_mask, struct spwlib_lis
 	 */
 	count = sizeof(buf) | GRSPW_READ_RECLAIM |
 		((dma_chan_mask << GRSPW_READ_CHANMSK_BIT) & GRSPW_READ_CHANMSK);
-	while ( (len=read(dev->fd, &buf[0], count)) > 0 ) {
+	while ( (len=READ(dev->fd, &buf[0], count)) > 0 ) {
 
 		/* Get number of Packets descriptorions copied */
 		len = len/sizeof(struct grspw_rtxpkt);
@@ -484,7 +557,7 @@ int grspw_tx_reclaim(struct grspw_dev *dev, int dma_chan_mask, struct spwlib_lis
 			pkt->flags = buf[i].flags;
 
 			/* Check Flags for errors */
-			if ( pkt->flags & TXPKT_FLAG_OUTPUT_MASK ) {
+			if ( pkt->flags & TXPKT_FLAG_LINKERR ) {
 				/* Something bad happended. Handle this somewhere
 				 * else? Where packets are handled.
 				 */
@@ -547,7 +620,7 @@ int grspw_rx_prepare(struct grspw_dev *dev, int dma_chan, struct spwlib_list *ls
 			count = (i+1)*sizeof(struct grspw_wrxpkt);
 		}
 		count |= (dma_chan << GRSPW_WRITE_CHAN_BIT) & GRSPW_WRITE_CHAN;
-		result = write(dev->fd, buf, count);
+		result = WRITE(dev->fd, buf, count);
 		if ( result < 0 ) {
 			PRINT_ERR("SpW%d: Prepare %d failed: %d\n", dev->index, spwlib_list_cnt(lst), errno);
 			return -1;
@@ -578,13 +651,13 @@ int grspw_rx_recv(struct grspw_dev *dev, int dma_chan_mask, struct spwlib_list *
 
 	tot = 0;
 	count = sizeof(buf) | ((dma_chan_mask << GRSPW_READ_CHANMSK_BIT) & GRSPW_READ_CHANMSK);
-	while ( (len=read(dev->fd, buf, count)) > 0 ) {
+	while ( (len=READ(dev->fd, buf, count)) > 0 ) {
 		len = len/sizeof(struct grspw_rrxpkt);
 		for (i=0; i<len; i++) {
 #if 0
 			/* Check driver internal problem... */
 			if ( buf[i].pkt_id >= dev->pkt_id_max ) {
-				PRINT_ERR("TX[%d] RECLAIM: Invalid Packet ID\n", dev->index);
+				PRINT_ERR("RX[%d] RECLAIM: Invalid Packet ID\n", dev->index);
 				return -1;
 			}
 #endif
@@ -595,6 +668,7 @@ int grspw_rx_recv(struct grspw_dev *dev, int dma_chan_mask, struct spwlib_list *
 			/* Check Flags for errors */
 			if ( pkt->flags & (RXPKT_FLAG_TRUNK|RXPKT_FLAG_EEOP) ) {
 				/* Something bad happended */
+				PRINT_ERR("RX[%d] RECV: pkt->flags error: 0x%08x\n", dev->index, pkt->flags);
 				return -1;
 			}
 
@@ -617,3 +691,57 @@ int grspw_rx_recv(struct grspw_dev *dev, int dma_chan_mask, struct spwlib_list *
 
 	return 0;
 }
+
+int grspw_tx_wait(struct grspw_dev *dev, int dma_chan, int send_cnt, int op, int sent_cnt, long timeout_ms)
+{
+	int result;
+	struct grspw_tx_wait_chan wait;
+
+	wait.chan = dma_chan;
+	wait.reserved = 0;
+	wait.timeout_ms = timeout_ms;
+	wait.send_cnt = send_cnt;
+	wait.op = op;
+	wait.sent_cnt = sent_cnt;
+
+	result = ioctl(dev->fd, GRSPW_IOCTL_TX_WAIT, &wait);
+	if ( result == -ETIME) {
+		return 2;
+	} else if ( result == -EIO) {
+		return 1;
+	} else if ( result == -EBUSY) {
+		return -1;
+	} else if ( result < 0 ) {
+		PRINT_ERR("Failed to TX-wait for dev %d.%d, errno: %d\n", dev->index, dma_chan, errno);
+		return -1;
+	}
+
+	return 0;
+}
+
+int grspw_rx_wait(struct grspw_dev *dev, int dma_chan, int recv_cnt, int op, int ready_cnt, long timeout_ms)
+{
+	int result;
+	struct grspw_rx_wait_chan wait;
+
+	wait.chan = dma_chan;
+	wait.reserved = 0;
+	wait.timeout_ms = timeout_ms;
+	wait.recv_cnt = recv_cnt;
+	wait.op = op;
+	wait.ready_cnt = ready_cnt;
+
+	result = ioctl(dev->fd, GRSPW_IOCTL_RX_WAIT, &wait);
+	if ( result == -ETIME) {
+		return 2;
+	} else if ( result == -EIO) {
+		return 1;
+	} else if ( result == -EBUSY) {
+		return -1;
+	} else if ( result < 0 ) {
+		PRINT_ERR("Failed to RX-wait for dev %d.%d, errno: %d\n", dev->index, dma_chan, errno);
+		return -1;
+	}
+
+	return 0;
+}
