@@ -1,7 +1,7 @@
 /*
- * GRSPW DMA SpaceWire packet tdemonstration software
+ * GRSPW DMA SpaceWire packet demonstration software
  *
- * Copyright (c) 2010 Aeroflex Gaisler AB
+ * Copyright (c) 2016 Cobham Gaisler AB
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -34,8 +34,7 @@
 #include "spwlib.h"
 #include "grspwlib.h"
 #include "maplib_user.h"
-
-#undef DEBUG_TEST
+#include "test.h"
 
 #ifdef DEBUG_TEST
 #define PRINT(str...) printf(str); fflush(NULL)
@@ -47,9 +46,6 @@
 #define PRINT3(str...)
 #endif
 
-#define DMACHAN 1
-#define DMACHAN_MASK (1<<DMACHAN)
-
 /******** MAP LIBRARY ********/
 #define MMAPLIB_POOL 0
 int map_alloc_pkt_idx = 0;
@@ -57,49 +53,6 @@ int map_alloc_pkt_ofs = 0;
 int map_alloc_hdr_idx = 0;
 int map_alloc_hdr_ofs = 0;
 
-/******** SpaceWire  ********/
-#define TXPKTS 5461 /* Max packet structures for 128Kb */
-#define RXPKTS 5461 /* Max packet structures for 128Kb */
-#define SPW_PROT_ID 155
-
-/* Number of seconds to run Test */
-/*#define TEST_LENGTH_SEC (60*60*12)*/
-#define TEST_LENGTH_SEC (30)
-
-/* Number of micro seconds to wait for all packets to be received */
-#define TEST_END_WAIT_MSEC 10000
-
-/* Skip RX check to save CPU and thereby get higher throughput figures
- * all packets are checked in the end any way...
- */
-#define SKIP_CHECK_FOR_EVERY_RECEIVED_PACKET
-
-/* SpaceWire Packet first sender (INITIATOR) device Index */
-#define SPW_INITIATOR 0
-
-/* SpW links on SpaceWire Router determine which PATH addess the AMBA
- * targets has
- */
-#define ROUTER_SPWLINK_NO 8
-
-/* Number of SpaceWire Links to share the packet buffers with, each
- * SpaceWire Link have 2 Packet Queues per DMA Unit.
- */
-#define SPW_LINKS_USED 4
-#define SPW_DMA_CHANS_PER_LINK_USED 1 /* only 1 supported at this time */
-#define SPW_DIRECTIONS_USED (SPW_LINKS_USED*SPW_DMA_CHANS_PER_LINK_USED*2)
-
-/*#define PKT_CNT 2016*/
-#define SPW_BUF_MB_BLOCKS 32
-#define SPW_BUF_SIZE (SPW_BUF_MB_BLOCKS*1024*1024)
-#define HDR_SIZE 16
-#ifndef PKT_SIZE
-  #define PKT_SIZE 1024
-#endif
-#define PKT_DATA_SIZE (PKT_SIZE-1-1-1-1-4) /* - sizeof(pkt_hdr)... */
-#define PKT_CNT  SPW_BUF_SIZE/PKT_SIZE/SPW_DIRECTIONS_USED
-#define PKT_CNT_TOT PKT_CNT*SPW_DIRECTIONS_USED
-
 /* Make PKT_CNT packets per RX/TX channel, each packet has a 1024 Bytes 
  * large buffer. Approx 8MBytes are used with PKT_CNT=2016.
  */
@@ -125,53 +78,14 @@ struct route_entry {
 	unsigned char dstadr[16];	/* 0 terminates array */
 };
 
-#if 0
-struct route_entry initial_route = 
-{
-	.dstadr_next = 1,
-	.dstadr = {2,0},
-};
-
-#define ROUTE_MAX 256
-struct route_entry routetab[ROUTE_MAX] = 
-{
-	{0, {0}},
-	{2, {1, 0}},
-	{1, {2, 0}},
-};
-#endif
-
-#if (SPW_LINKS_USED == 2)
-struct route_entry initial_route = 
-{
-	.dstadr_next = 0x1,
-	.dstadr = {1+ROUTER_SPWLINK_NO+1, 0x2, 0},
-};
-
-#define ROUTE_MAX 256
-struct route_entry routetab[ROUTE_MAX] = 
-{
-	{0, {0, 0}},
-	{0x2, {1+ROUTER_SPWLINK_NO, 0x1, 0}},
-	{0x1, {1+ROUTER_SPWLINK_NO+1, 0x2, 0}},
-};
-#elif (SPW_LINKS_USED == 4)
-struct route_entry initial_route = 
-{
-	.dstadr_next = 0x3,
-	.dstadr = {1+ROUTER_SPWLINK_NO+1, 0x2, 0},
-};
-
-#define ROUTE_MAX 256
-struct route_entry routetab[ROUTE_MAX] = 
-{
-	/* DST */  /* NEXT-ADR, {PATH/LOGICAL ADDRESS} */
-	/* 0x00 */ {0, {0, 0}},
-	/* 0x01 */ {0x2, {1+ROUTER_SPWLINK_NO, 0x1, 0}},
-	/* 0x02 */ {0x3, {1+ROUTER_SPWLINK_NO+1, 0x2, 0}},
-	/* 0x03 */ {0x4, {1+ROUTER_SPWLINK_NO+2, 0x3, 0}},
-	/* 0x04 */ {0x1, {1+ROUTER_SPWLINK_NO+3, 0x4, 0}},
-};
+#if !defined(ROUTEINIT) || (ROUTEINIT == 0)
+ #include "routeinit_auto_router.c"
+#elif (ROUTEINIT == 1)
+ #include "routeinit_custom_router.c"
+#elif (ROUTEINIT == 2)
+ #include "routeinit_dual.c"
+#else
+ #error INVALID ROUTE INIT CONFIGURATION FILE
 #endif
 
 /* Packet ID -> Packet Pointer. All Devices share the Packet ID. */
@@ -191,8 +105,10 @@ struct grspw_device {
 	struct spwlib_list tx_list;
 	struct spwlib_list check_list;
 	struct route_entry *routetab;
-	int pkt_number_next;
-	int check_no;
+	int rxpkt_number_next;
+	int txpkt_number_next;
+	int rxcheck_no;
+	int txcheck_no;
 };
 struct grspw_device devs[8];
 
@@ -203,7 +119,7 @@ struct grspw_config dev_configs[8] =
 		.adrcfg =
 		{
 			.promiscuous = 0,
-			.def_addr = 0x1,
+			.def_addr = 0x1, /* Node addr. */
 			.def_mask = 0,
 			.dma_nacfg =
 			{
@@ -419,6 +335,8 @@ int grspw_init(int idx)
 	struct grspw_dev *dev = &device->dev;
 	struct grspw_link_state ls;
 	struct grspw_link_ctrl lctrl;
+	unsigned int status;
+	int time;
 
 	memset(device, 0, sizeof(struct grspw_device));
 	if ( grspw_open(dev, idx) ) {
@@ -471,6 +389,7 @@ int grspw_init(int idx)
 	lctrl.ctrl = LINKOPTS_ENABLE | LINKOPTS_AUTOSTART | LINKOPTS_START;
 	lctrl.clkdiv_start = 2;
 	lctrl.clkdiv_run = 0;
+	lctrl.stscfg = LINKSTS_MASK; /* let driver handle link errors */
 	if ( grspw_link_set(dev, &lctrl) != 0 ) {
 		printf("Succeeded to select port1, however only one PORT on dev %d!\n", dev->index);
 		return -1;
@@ -483,11 +402,30 @@ int grspw_init(int idx)
 	}
 	grspw_linkstate_print(&ls);
 
+	status = 0xffffffff; /* anything.. just debug */
+	if (grspw_status_read(dev, &status)) {
+		printf("Failed to read the status register for dev %d\n", dev->index);
+		return -1;
+	}
+	printf("GRSPW STATUS Register: 0x%08x\n", status);
+
 	if ( (dev->hwsup.hw_version >> 16) == GAISLER_SPW2_DMA ) {
 		printf("\n\n###NOTE: running on SPW-ROUTER DMA SpaceWire link:\n");
 		printf(    "         there is no Link-state available\n");
 	}
 
+	/* Set TIMECNT and TCTRL to 0. Do not generate Tick-In */
+	if (grspw_tc_set(dev, GRSPW_TC_SEND_OPTION_SET_TCTRL | 0x00)) {
+		printf("Failed to set TCTRL|TIMECNT for dev %d\n", dev->index);
+		return -1;
+	}
+	/* Check that it is zero in register.. */
+	time = 1; /* anything.. just for debug */
+	if (grspw_tc_read(dev, &time) || time != 0x00) {
+		printf("Failed to get TCTRL|TIMECNT for dev %d TIME=%d\n", dev->index, time);
+		return -1;
+	}
+
 	if ( grspw_stats_clear(dev) ) {
 		return -1;
 	}
@@ -506,91 +444,6 @@ void grspw_cleanup(int idx)
 	}
 }
 
-#if 0
-/* Memory MAP memory to use for example */
-int mmaplib_init(int idx, int blk_size, int blk_cnt)
-{
-	int i;
-	void *adr;
-	unsigned int start, end, len;
-	struct maplib_mmap_info *mapi = &mmapinfo;
-	char devname[32];
-	struct maplib_setup setup;
-
-	sprintf(devname, "/dev/maplib%d", idx);
-	map_fd = open(devname, O_RDWR);
-	if ( map_fd < 0 ) {
-		printf("Failed to open MMAPLib\n");
-		return -1;
-	}
-
-	/* Setup MAP */
-	setup.blk_size = blk_size;
-	setup.blk_cnt = blk_cnt;
-	if ( ioctl(map_fd, MAPLIB_IOCTL_SETUP, &setup) ) {
-		printf("Failed to Setup MMAPLib: %d\n", errno);
-		return -2;
-	}
-
-	/* Get MMAP information calculated by driver */
-	if ( ioctl(map_fd, MAPLIB_IOCTL_MMAPINFO, mapi) ) {
-		printf("Failed to get MMAPINFO, errno: %d\n", errno);
-		return -3;
-	}
-
-	printf("MMAP INFO: GRSPW Packet Buffers:\n");
-	printf("    offset: 0x%x\n", mapi->buf_offset);
-	printf("    length: %d (0x%x)\n", mapi->buf_length, mapi->buf_length);
-	printf("    MMAP Block size: 0x%x\n", mapi->buf_blk_size);
-
-	memset(maps, 0, sizeof(maps));
-
-	/* Map all SpaceWire Packet Buffers */
-	start = mapi->buf_offset;
-	end = mapi->buf_offset + mapi->buf_length;
-
-	/* Memory MAP driver's Buffers READ-and-WRITE  */
-	adr = mmap(NULL, mapi->buf_length, PROT_READ|PROT_WRITE, MAP_SHARED,
-			map_fd, start);
-	if ( (unsigned int)adr == 0xffffffff ) {
-		printf("MMAP Bufs Failed: %p, errno %d, %x\n", adr, errno, mapi->buf_length);
-		return -4;
-	}
-	i=0;
-	while ( start < end ) {
-		/* Calc length of MMAP Block */
-		if ( (end - start) >= mapi->buf_blk_size )
-			len = mapi->buf_blk_size;
-		else
-			len = end - start;
-
-		maps[i].start = adr;
-		maps[i].length = len;
-		maps[i].ofs = start;
-		/*printf("MAPPED OFS=%p, BASE=%x, LEN=%p\n", maps[i].ofs, maps[i].start, maps[i].length);*/
-		i++;
-
-		/* Next Block */
-		start += mapi->buf_blk_size;
-		adr += len;
-	}
-
-	printf("M-Mapping successful\n");
-
-	return 0;
-}
-
-int mmaplib_cleanup(void)
-{
-	/* Close MAP Lib, this will automatically unmap all memory for us. */
-	close(map_fd);
-
-	maps[0].length = 0;
-	maps[0].start = 0;
-	maps[0].ofs = 0;
-}
-#endif
-
 /* It is important that block boundaries are not overlapped in memory alllcated
  * here.
  */
@@ -602,9 +455,6 @@ void *spw_alloc_pkt(void *data, int poolno, int hdr, int length)
 	int *map_alloc_ofs;
 
 	if ( hdr ) {
-		/* It is assumed that the header will never fill a MAP 
-		 * block, so the range check (map->length==0) is not
-		 * completely correcy... however will work. */
 		map_alloc_idx = &map_alloc_hdr_idx;
 		map_alloc_ofs = &map_alloc_hdr_ofs;
 	} else {
@@ -620,24 +470,27 @@ try:
 		*map_alloc_ofs = 0;
 		goto try;
 	} else {
+		if ((sizeof(maps) / sizeof(maps[0])) <= *map_alloc_idx) {
+			printf("ERROR: map[%d] referenced\n", *map_alloc_idx);
+			return NULL;
+		}
 		retval = (void *)((unsigned int)map->start + *map_alloc_ofs);
 		(*map_alloc_ofs) += length;
 	}
-	/*printf("ALLOCED: %p %d\n", retval, length);*/
+	/*printf("ALLOCED %s: %p %d\n", hdr ? "hdr" : "data", retval, length);*/
 	return retval;
 }
 
 void spw_alloc_reset(void *data)
 {
 #if HDR_SIZE > 0
-	/* Let headers have first MAP block */
-	map_alloc_pkt_idx = 1;
-	map_alloc_pkt_ofs = 0;
+	/* Let headers have the first MAP blocks */
+	map_alloc_pkt_idx = HDR_BUF_128K_BLOCKS;
 #else
 	/* No headers: Let headers have NO MAP block */
 	map_alloc_pkt_idx = 0;
-	map_alloc_pkt_ofs = 0;
 #endif
+	map_alloc_pkt_ofs = 0;
 
 	/* Let Header start at index 0. */
 	map_alloc_hdr_idx = 0;
@@ -725,7 +578,7 @@ int pkts_init(struct grspw_device *device, int initiator)
 	if ( initiator ) {
 		/* Initiator */
 
-		/* Allocate PKT_CNT 1k Packets for RX/TX buffers */
+		/* Allocate PKT_CNT packets with sizes PKT_SIZE, HDR_SIZE for RX/TX buffers */
 		cnt = spwlib_pkt_chain_take(poolset, PKT_SIZE, HDR_SIZE, PKT_CNT, &device->rx_prep_list);
 		if ( cnt != PKT_CNT ) {
 			printf("Failed allocating RX packets for %d\n", initiator);
@@ -863,7 +716,7 @@ int pktlist_check_rx(struct grspw_device *device, struct spwlib_list *lst, int c
 	unsigned short exp, got;
 	int i;
 
-	device->check_no++;
+	device->rxcheck_no++;
 	while ( pkt ) {
 		/* Check Reception HW-Flags */
 		if ( (pkt->flags & (RXPKT_FLAG_TRUNK|RXPKT_FLAG_EEOP|RXPKT_FLAG_RX))
@@ -874,9 +727,9 @@ int pktlist_check_rx(struct grspw_device *device, struct spwlib_list *lst, int c
 
 		/* Check Packet number (Sequence check) */
 		pkt_hdr = (struct pkt_hdr *)pkt->data;
-		if ( pkt_hdr->pkt_number != device->pkt_number_next ) {
+		if ( pkt_hdr->pkt_number != device->rxpkt_number_next ) {
 			printf("PKT RX SEQ ERROR: exp %d, got %d\n",
-				device->pkt_number_next, pkt_hdr->pkt_number);
+				device->rxpkt_number_next, pkt_hdr->pkt_number);
 			return -2;
 		}
 		/* Check Packet Content */
@@ -886,7 +739,7 @@ int pktlist_check_rx(struct grspw_device *device, struct spwlib_list *lst, int c
 				got = pkt_hdr->data[i];
 				if ( exp != got ) {
 					debug_test(got, exp, &pkt_hdr->data[i]);
-					printf(" BAD DATA in PACKET: got 0x%04x expected 0x%04x (%p)\n",
+					printf(" BAD RX DATA in PACKET: got 0x%04x expected 0x%04x (%p)\n",
 						got, exp, &pkt_hdr->data[i]);
 					return -1;
 				}
@@ -894,14 +747,52 @@ int pktlist_check_rx(struct grspw_device *device, struct spwlib_list *lst, int c
 			}
 		}
 		/* Prepare next Packet content */
-		device->pkt_number_next++;
-		if (device->pkt_number_next >= PKT_CNT)
-			device->pkt_number_next = 0;
-#if 0
-		if ( device->dev.index == 1 ) {
-			huge_bus_ofs += sprintf(&huge_buf[huge_bus_ofs], " %03d: FLG=%x,NUM=%d\n", device->check_no, pkt->flags, pkt_hdr->pkt_number);
+		device->rxpkt_number_next++;
+		if (device->rxpkt_number_next >= PKT_CNT)
+			device->rxpkt_number_next = 0;
+
+		pkt = pkt->next;
+	}
+	return 0;
+}
+
+int pktlist_check_tx(struct grspw_device *device, struct spwlib_list *lst, int content)
+{
+	struct spwlib_pkt *pkt = lst->head;
+	struct pkt_hdr *pkt_hdr;
+	unsigned short exp, got;
+	int i;
+
+	device->txcheck_no++;
+	while ( pkt ) {
+		/* Check Packet number (Sequence check) */
+		pkt_hdr = (struct pkt_hdr *)pkt->data;
+		if ( pkt_hdr->pkt_number != device->txpkt_number_next ) {
+			printf("PKT TX SEQ ERROR: exp %d, got %d\n",
+				device->txpkt_number_next, pkt_hdr->pkt_number);
+			return -2;
 		}
-#endif
+
+		/* Check Packet Content */
+		if ( content ) {
+			exp = pkts_content_tab[pkt_hdr->pkt_number];
+			for (i=0; i<PKT_DATA_SIZE/2; i++) {
+				got = pkt_hdr->data[i];
+				if ( exp != got ) {
+					debug_test(got, exp, &pkt_hdr->data[i]);
+					printf(" BAD TX DATA in PACKET: got 0x%04x expected 0x%04x (%p)\n",
+						got, exp, &pkt_hdr->data[i]);
+					return -1;
+				}
+				exp++;
+			}
+		}
+
+		/* Prepare next Packet content */
+		device->txpkt_number_next++;
+		if (device->txpkt_number_next >= PKT_CNT)
+			device->txpkt_number_next = 0;
+
 		pkt = pkt->next;
 	}
 	return 0;
@@ -1005,19 +896,38 @@ int grspw_process(struct grspw_device *device, int options, int stop)
 	PRINT("5A\n");
 	if ( (options & 2) && spwlib_list_is_empty(&device->tx_list) == 0 ) {
 		PRINT("5B\n");
+
+#ifndef SKIP_CHECK_FOR_EVERY_TRANSMITTED_PACKET
+		/* Check Packet Sequence Number and HW-Transmission Flags */
+		if ( pktlist_check_tx(device, &device->tx_list, 0) ) {
+			printf(" ERROR In TX Packet sequence\n");
+			return -17;
+		}
+#endif
 		result = grspw_tx_send(dev, DMACHAN, &device->tx_list);
 		if ( result < 0 ) {
 			printf("Send failed: %d (%d)\n", result, dev->index);
 			return -20;
 		}
-		spwlib_list_clr(&device->tx_list);
+#ifndef SKIP_CHECK_FOR_EVERY_TRANSMITTED_PACKET
+		if ( spwlib_list_is_empty(&device->tx_list) == 0 ) {
+			printf(" The TX-check only works when GRSPW driver has "
+			       "so many frame structures so that all TX "
+			       "packets are accepted every time. Please turn it "
+			       "off by the define.");
+			exit(-1);
+		}
+#endif
 	}
 
 	return 0;
 }
 
-int main(void)
+int main(int argc, char *argv[])
 {
+	int nlinks = ROUTER_SPWLINK_NO;
+	int usage_ok = 0;
+
 	int i, result, shutdown, stop, devno, options, tot;
 	struct timeval t0, t1, t2;
 	double time;
@@ -1026,12 +936,29 @@ int main(void)
 	struct grspw_stats stats;
 	struct grspw_qpktcnt qpktcnt;
 
+	if (1 == argc) {
+		usage_ok = 1;
+	} else if (2 == argc) {
+		char *opt = argv[1];
+		if ('-' == opt[0] && 's' == opt[1]) {
+			nlinks = atoi(opt+2);
+			usage_ok = 1;
+		}
+	}
+
+	if (!usage_ok) {
+		printf("Usage: %s -s[spw_links]\n", argv[0]);
+		return 1;
+	}
+
+	init_routetab(nlinks);
+
 	for (i=0; i<SPW_LINKS_USED; i++)
 		devs[i].dev.fd = -1;
 
 	printf("Test built for SpW-router with %d SpW Ports and at least %d AMBA ports.\n"
 	       "Note that the SpW-links will not be used.\n",
-	       ROUTER_SPWLINK_NO, SPW_LINKS_USED);
+	       nlinks, SPW_LINKS_USED);
 	printf("\n\n");
 
 	/* Allocate SPW_BUF_MB_BLOCKS Mbyte Memory mapped memory in blocks
@@ -1039,7 +966,7 @@ int main(void)
 	 * One extra block is allocated for header Data (a of waste of memory..)
 	 */
 #if HDR_SIZE > 0
-	hdrblk = 1;
+	hdrblk = HDR_BUF_128K_BLOCKS;
 #endif
 	if ( mmaplib_init(MMAPLIB_POOL, 128*1024, 8*SPW_BUF_MB_BLOCKS+hdrblk) ) {
 		return 1;
@@ -1067,6 +994,8 @@ int main(void)
 	printf("PKT_CNT:             %d\n", PKT_CNT);
 	printf("PKT_CNT_TOT:         %d\n", PKT_CNT_TOT);
 	printf("PKT_SIZE:            %d\n", PKT_SIZE);
+	printf("HDR_BUF_SIZE:        %d\n", HDR_BUF_SIZE);
+	printf("HDR_BUF_128K_BLOCKS: %d\n", HDR_BUF_128K_BLOCKS);
 	printf("SPW_LINKS_USED:      %d\n", SPW_LINKS_USED);
 	printf("DMA_CHANS_PER_LINK:  %d\n", SPW_DMA_CHANS_PER_LINK_USED);
 	printf("DIRECTIONS:          %d\n", SPW_DIRECTIONS_USED);
@@ -1099,8 +1028,10 @@ int main(void)
 		devs[i].routetab = &routetab[0];
 
 		/* Init Packet Check */
-		devs[i].pkt_number_next = 0;
-		devs[i].check_no = 0;
+		devs[i].rxpkt_number_next = 0;
+		devs[i].txpkt_number_next = 0;
+		devs[i].rxcheck_no = 0;
+		devs[i].txcheck_no = 0;
 	}
 
 	printf("\n\n");
@@ -1234,9 +1165,11 @@ int main(void)
 		printf(" DRVQ RX_READY: %d\n", qpktcnt.chan[0].rx_ready);
 		printf(" DRVQ RX_SCHED: %d\n", qpktcnt.chan[0].rx_sched);
 		printf(" DRVQ RX_RECV: %d\n", qpktcnt.chan[0].rx_recv);
+		printf(" DRVQ RX_HWCNT: %d\n", qpktcnt.chan[0].rx_hw);
 		printf(" DRVQ TX_SEND: %d\n", qpktcnt.chan[0].tx_send);
 		printf(" DRVQ TX_SCHED: %d\n", qpktcnt.chan[0].tx_sched);
 		printf(" DRVQ TX_SENT: %d\n", qpktcnt.chan[0].tx_sent);
+		printf(" DRVQ TX_HWCNT: %d\n", qpktcnt.chan[0].tx_hw);
 
 		grspw_stats_get(&devs[i].dev, &stats);
 		grspw_stats_print(&devs[i].dev, &stats);
@@ -1261,7 +1194,7 @@ int main(void)
 	if ( spwlib_list_is_empty(&devs[SPW_INITIATOR].check_list) == 0 ) {
 		struct pkt_hdr *pkt_hdr = (struct pkt_hdr *)
 			devs[SPW_INITIATOR].check_list.head->data;
-		devs[SPW_INITIATOR].pkt_number_next = pkt_hdr->pkt_number;
+		devs[SPW_INITIATOR].rxpkt_number_next = pkt_hdr->pkt_number;
 
 		result = pktlist_check_rx(&devs[SPW_INITIATOR],
 					&devs[SPW_INITIATOR].check_list,
